@@ -9,6 +9,7 @@ import time
 import random
 from functools import lru_cache
 import tqdm
+from datetime import datetime
 from . import util
 # hidden import below: win32com (pip install pywin32) only needed for .compact_db if using Access db.
 
@@ -19,14 +20,13 @@ class DataBase():
 
     def __init__(self, db_location_str: str='', create_if_none: bool=True) -> None:
         self.db_location_str = db_location_str
-        self._pull_table_cache: dict = {}
+        self._pull_cache: dict = {}
 
         self.db_type = self._find_db_type()
         if self.db_type == 'ACCESS':
             self.connection = self._connection_access
             try:
-                conn = self.connection()
-                conn.close()
+                self.connection()
             except pyodbc.Error as error:
                 print(error)
                 print(f'\nERROR with pyodbc!  Unable to connect to Access Database: {self.db_location_str}')
@@ -45,9 +45,9 @@ class DataBase():
 
 
 
-    def _find_db_type(self):
+    def _find_db_type(self) -> str:
         '''
-        Figure out what kind of databse is being used.
+        Figure out what kind of database is being used.
         '''
         if self.db_location_str in os.environ:  # Environment Variables are case-insensitive
             print(f'{self.db_location_str} found as Environment Variable.  Substituting database path.')
@@ -158,12 +158,7 @@ class DataBase():
             print(f'Current database is: {self.db_type}')
 
 
-    def pull(self, *args, **kwargs) -> list:
-        '''Shorthand for pull_table'''
-        return self.pull_table(*args, **kwargs)
-
-
-    def pull_table(self, tablename: str, columns='all', clear_cache=False, progress_handler=None) -> list:
+    def pull(self, tablename: str, columns='all', fresh=False, progress_handler=None) -> list:
         '''
         "SELECT *" query for full table as specified from tablename.
         ALSO WORKS for an Access Select query named tablename!
@@ -172,8 +167,8 @@ class DataBase():
         to pull the full table for ONLY those columns.
 
         NOTE!  This function uses caching to avoid extra queries for the same data.
-        clear_cache kwarg provides ability to clear cache and pull data
-        with a fresh query.  Set clear_cache=True in the event that the database
+        "fresh" kwarg provides ability to clear cache and pull data
+        with a fresh query.  Set fresh=True in the event that the database
         table may have been updated since any previous calls.
 
         progress_handler kwarg can be used to provide status updates to a callback.
@@ -189,9 +184,9 @@ class DataBase():
             print(f'Table or query "{tablename}" not found.  Pull aborted.')
             return []
 
-        if clear_cache:
-            self._pull_table_cache = {}
-            return self.pull_table(tablename, columns)
+        if fresh:
+            self._pull_cache = {}
+            return self.pull(tablename, columns)
         else:
             # check for questionable table/column names
             for name in [tablename] + list(columns):
@@ -199,7 +194,7 @@ class DataBase():
                     return []
 
             requested_data_key = f'{tablename}_' + '_'.join(sorted(columns))  # key string for caching db pulls in dict
-            if requested_data_key not in self._pull_table_cache:
+            if requested_data_key not in self._pull_cache:
                 if columns == 'all':
                     sql = f'SELECT * FROM "{tablename}";'
                 elif isinstance(columns, str):
@@ -214,30 +209,38 @@ class DataBase():
                     else:
                         print('progress_handler is only available for use with a SQLite database.')
 
-                self._pull_table_cache[requested_data_key] = util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type)
+                self._pull_cache[requested_data_key] = util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type)
                 conn.close()
-            return self._pull_table_cache[requested_data_key]
+            return self._pull_cache[requested_data_key]
 
 
-    def pull_table_where(self, tablename: str, condition: str) -> list:
+    def pull_where(self, tablename: str, condition: str, columns='all') -> list:
         '''
         SELECT * WHERE Query for table as specified from tablename and condition
         Return list of dicts for rows with column names as keys.
         '''
-        sql = f'SELECT * FROM {tablename} WHERE {condition};'
+        if columns == 'all':
+            sql = f'SELECT * FROM {tablename} WHERE {condition};'
+        elif isinstance(columns, list):  # list of columns to pull
+            sql = f'SELECT {", ".join(columns)} FROM {tablename} WHERE {condition};'
+        else:
+            print('Columns kwarg for .pull_where must be a list of column names.')
         conn, cursor = self.connection(also_cursor=True)
         data = util.list_of_dicts_from_query(cursor, sql, tablename, self.db_type)
         conn.close()
         return data
 
 
-    def pull_table_where_id_in_list(self, tablename: str, id_col: str, match_values: list, use_multip: bool=False, progressbar: bool=True) -> list:
+    def pull_where_id_in_list(self, tablename: str, id_col: str, match_values: list, columns='all', use_multip: bool=False, progressbar: bool=True) -> list:
         '''
         Pulls all data from table where id_col value is in the provided match_values.
         '''
         @lru_cache(maxsize=4)
         def sql_str(subset_len: int) -> str:
-            return f"SELECT * FROM [{tablename}] WHERE {id_col} in ({'?,'.join(['' for _ in range(subset_len)])}?);"
+            if columns == 'all':
+                return f"SELECT * FROM [{tablename}] WHERE {id_col} in ({'?,'.join(['' for _ in range(subset_len)])}?);"
+            else:
+                return f"SELECT {', '.join(columns)} FROM [{tablename}] WHERE {id_col} in ({'?,'.join(['' for _ in range(subset_len)])}?);"
 
         if progressbar:
             pbar = tqdm.tqdm(total=len(match_values))
@@ -259,6 +262,7 @@ class DataBase():
         return data
 
 
+    @lru_cache(maxsize=1)
     def table_names(self) -> list:
         '''
         Return sorted list of all tables in the database.
@@ -273,6 +277,7 @@ class DataBase():
         return sorted(tables)
 
 
+    @lru_cache(maxsize=1)
     def query_names(self) -> list:
         '''
         Return sorted list of all queries in the database.
@@ -285,6 +290,7 @@ class DataBase():
             return []
 
 
+    @lru_cache(maxsize=64)
     def columns_and_types(self, tablename: str) -> dict:
         '''
         Return dict of all column: type pairs in specified table.
@@ -320,11 +326,11 @@ class DataBase():
         if not self.db_type == 'ACCESS':
             print('ERROR!  .key_columns is only currently implemented for Access databases.')
         with self as cursor:
-            key_cols = [row[8] for row in cursor.statistics(tablename) if row[5] == 'PrimaryKey']
+            key_cols = [row[8] for row in cursor.statistics(tablename) if row[5] and 'key' in row[5].lower()]
         return key_cols
 
 
-    def create_table(self, tablename: str, columns_and_types: dict, force_overwrite: bool=False):
+    def create_table(self, tablename: str, columns_and_types: dict, force_overwrite: bool=False) -> None:
         '''
         Create a table in the database with name "tablename"
 
@@ -374,6 +380,8 @@ class DataBase():
         conn.close()
         if create_complete:
             print(f'Table {tablename} successfully created.')
+            self.table_names.cache_clear()  # need to repull table names as just created a new one
+            self.columns_and_types.cache_clear()
         else:
             print(error)
             print(f'\nUnable to create table "{tablename}"\nPerhaps the database is locked?!')
@@ -389,7 +397,7 @@ class DataBase():
         if not keys:
             return set()
         else:
-            existing_tups = set(tuple(row[key] for key in keys) for row in self.pull(tablename, columns=keys, clear_cache=True))
+            existing_tups = set(tuple(row[key] for key in keys) for row in self.pull(tablename, columns=keys, fresh=True))
             dups = set()
             for d in data:
                 tup = tuple(d[key] for key in keys)
@@ -398,13 +406,18 @@ class DataBase():
             return dups
 
 
-    def append(self, tablename: str, data: Union[List[dict], dict], create_table_if_needed: bool=True, safe=False, clean_column_names=False, robust: bool=True):
+    def append(self, tablename: str, data: Union[List[dict], dict], create_table_if_needed: bool=True, safe=False, clean_column_names=False, robust: bool=True) -> None:
         '''
         Append rows of data to database table.
         Create the table in the database if it doesn't exist if create_table_if_needed is True
 
         "data" arg is list of row dicts where each row dict contains all columns as keys.
 
+        "safe" kwarg is False by default and parameterized insert queries are used.
+        IF you know the data is ~~~VERY SAFE~~~ without a chance of SQL Injection...
+        ...then setting safe=True converts data into direct SQL strings for faster imports.
+        Note that safe=True is more delicate and sensitive to the quality of data.
+
         "robust" kwarg enables automatic data cleaning (type conversions, null, missing columns) if True.
         Setting robust to False improves speed if using clean input data.
         '''
@@ -417,17 +430,14 @@ class DataBase():
 
         if clean_column_names:
             print('Cleaning column names in data to be appended.')
-            data_keys = list(data[0].keys())
-            for row in data:
-                for key in data_keys:
+            for key in list(data[0].keys()):
+                for row in data:
                     row[util.clean_column_name(key)] = row.pop(key)
 
         if tablename not in self.table_names() and create_table_if_needed:
-            columns_and_types = {key: type(value).__name__ for key, value in data[0].items()}
-            self.create_table(tablename, columns_and_types)
+            self.create_table(tablename, {key: type(value).__name__ for key, value in data[0].items()})
         elif tablename not in self.table_names() and not create_table_if_needed:
-            print(f'ERROR!  Table "{tablename}" does not exist in database!')
-            print('Use create_table_if_needed=True if you would like to create it.')
+            print(f'ERROR!  Table "{tablename}" does not exist in database!\nUse create_table_if_needed=True if you would like to create it.')
             return None
 
         if robust:
@@ -464,6 +474,16 @@ class DataBase():
                 print(f'The remaining {len(non_dup_data)} rows are still being appended.\n')
                 data = non_dup_data
 
+        is_access = True if self.db_type == 'ACCESS' else False
+        def convert_to_sql(value):
+            if value is None:
+                return 'NULL'
+            elif isinstance(value, str):
+                return f"'{value}'"
+            elif isinstance(value, datetime) and is_access:
+                return f'#{value}#'  # adding "#" on either end makes the Access date insert works
+            else:
+                return value
 
         conn, cursor = self.connection(also_cursor=True)
         pbar = tqdm.tqdm(total=len(data))
@@ -472,8 +492,8 @@ class DataBase():
             try:
                 if safe:
                     for row in data[-100:]:
-                        cursor.execute(insert_sql + "(" + ','.join(["'" + str(row[col]) + "'" for col in columns]) + ");")
-                elif not safe:
+                        cursor.execute(insert_sql + '(' + ','.join([f'{convert_to_sql(row[col])}' for col in columns]) + ');')
+                else:
                     cursor.executemany(insert_many_sql, [tuple(row_dict[col] for col in columns) for row_dict in data[-100:]])
                 pbar.update(100 if len(data) >= 100 else len(data))
                 data = data[:-100]
@@ -483,11 +503,11 @@ class DataBase():
         pbar.close()
         conn.commit()
         conn.close()
-        self._pull_table_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something has been updated
         print(f'Data inserted in "{tablename}" -> {"{:,.0f}".format(original_data_len)} rows')
 
 
-    def update(self, tablename: str, match_col: str, match_val, update_col: str, update_val, progress_handler=None):
+    def update(self, tablename: str, match_col: str, match_val, update_col: str, update_val, progress_handler=None) -> None:
         '''
         Update a database table with a value or values.
 
@@ -539,36 +559,38 @@ class DataBase():
         else:
             cursor.execute(sql, (update_val, match_val))
         conn.commit()
-        self._pull_table_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
 
 
-    def add_column(self, tablename: str, new_col: str, new_type='str'):
+    def add_column(self, tablename: str, new_col: str, new_type='str') -> None:
         '''Add a new column to a database table.'''
         if new_col in self.columns_and_types(tablename):
             print(f'Column {new_col} is already in {tablename}!')
             return
         if new_type == 'str':
             new_type = 'varchar(255)' if self.db_type == 'ACCESS' else 'TEXT'
-        conn, cursor = self.connection(also_cursor=True)
-        cursor.execute(f'ALTER TABLE {tablename} ADD COLUMN {new_col} {new_type};')
-        conn.commit()
+        with self as cursor:
+            cursor.execute(f'ALTER TABLE {tablename} ADD COLUMN {new_col} {new_type};')
         print(f'Column {new_col} added to {tablename}.')
-        self._pull_table_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        self.columns_and_types.cache_clear()
 
 
-    def drop_column(self, tablename: str, column: str):
+    def drop_column(self, tablename: str, column: str) -> None:
         '''Remove a column from a database table.'''
         if column not in self.columns_and_types(tablename):
             print(f'Column {column} does not exist in {tablename}.')
             return
-        conn, cursor = self.connection(also_cursor=True)
-        cursor.execute(f'ALTER TABLE {tablename} DROP COLUMN "{column}";')
-        conn.commit()
+
+        with self as cursor:
+            cursor.execute(f'ALTER TABLE {tablename} DROP COLUMN "{column}";')
+
         print(f'Column {column} removed from {tablename}')
-        self._pull_table_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+        self.columns_and_types.cache_clear()
 
 
-    def delete_duplicates(self, tablename: str, grouping_columns=None):
+    def delete_duplicates(self, tablename: str, grouping_columns=None) -> None:
         '''
         Delete duplicate rows from a db table while retaining most recently added row.
         Duplicates are determined by grouping based on the grouping_columns kwarg (provide iterable).
@@ -585,10 +607,8 @@ class DataBase():
 
         if self.db_type == 'SQLITE':
             with self as cursor:
-            # conn, cursor = self.connection(also_cursor=True)
                 cursor.execute(f'DELETE FROM {tablename} WHERE rowid NOT IN (SELECT max(rowid) FROM {tablename} GROUP BY {", ".join(grouping_columns)})')
-            # conn.commit()
-            self._pull_table_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+            self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
 
         elif self.db_type == 'ACCESS':
             # TODO:  Think some sort of SQL can accomplish dup deletion better than in Python... haven't figured it out yet
@@ -616,25 +636,21 @@ class DataBase():
                     existing_combos.add(row_combo)
             with self as cursor:
                 cursor.execute(f'DELETE * FROM {tablename};')
-            self.append(tablename, list(reversed(new_data)))  # UN-reverse table entries
-            self._pull_table_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something's been updated
+            self.append(tablename, list(reversed(new_data)), safe=True, robust=False)  # UN-reverse table entries
+            self._pull_cache.pop(tablename, None)  # clear cache for this table as want new table pull if something has been updated
 
 
-    def create_index(self, tablename: str, column: str, index_name: str='', unique: bool=False):
+    def create_index(self, tablename: str, column: str, index_name: str='', unique: bool=False) -> None:
         if self.db_type == 'SQLITE':
             index_name = column if index_name == '' else index_name  # use column name if not provided
-            conn, cursor = self.connection(also_cursor=True)
-            if unique:
-                cursor.execute(f'CREATE UNIQUE INDEX {index_name} on {tablename}({column});')
-            else:
-                cursor.execute(f'CREATE INDEX {index_name} on {tablename}({column});')
-            conn.commit()
-            conn.close()
+            with self as cursor:
+                cursor.execute(f'CREATE {"UNIQUE " if unique else ""}INDEX {index_name} on {tablename}({column});')
+            self.columns_and_types.cache_clear()
         else:
             print('.create_index is currently only implemented for SQLite databases.')
 
 
-    def drop_table(self, tablename: str):
+    def drop_table(self, tablename: str) -> None:
         '''
         Drop/delete the specified table from the database.
         '''
@@ -644,34 +660,33 @@ class DataBase():
 
         if self.db_type == 'SQLITE':
             t0, drop_complete = time.time(), False
-            conn, cursor = self.connection(also_cursor=True)
-            while time.time() - t0 < 10:
-                try:
-                    cursor.execute(f'DROP TABLE IF EXISTS "{tablename}";')
-                    conn.commit()
-                    drop_complete = True
-                    break
-                except sqlite3.OperationalError:
-                    pass
-            conn.close()
+            with self as cursor:
+                while time.time() - t0 < 10:
+                    try:
+                        cursor.execute(f'DROP TABLE IF EXISTS "{tablename}";')
+                        drop_complete = True
+                        break
+                    except sqlite3.OperationalError:
+                        pass
+
             if drop_complete:
                 print(f'Table "{tablename}" deleted.')
             else:
                 print(f'Unable to drop table "{tablename}" as the database is locked!')
         elif self.db_type == 'ACCESS':
-            conn, cursor = self.connection(also_cursor=True)
-            cursor.execute(f'DROP TABLE {tablename};')
-            conn.commit()
-            conn.close()
+            with self as cursor:
+                cursor.execute(f'DROP TABLE {tablename};')
             print(f'Table {tablename} deleted.')
         else:
             print('ERROR!  Table deletion only implemented in SQLite and Access currently.')
             return
 
-        self._pull_table_cache.pop(tablename, None)  # clear cache for this table as table has been dropped
+        self._pull_cache.pop(tablename, None)  # clear cache for this table as table has been dropped
+        self.table_names.cache_clear()  # need to repull table names as just (likely) deleted one
+        self.columns_and_types.cache_clear()
 
 
-    def copy_table(self, other_db, tablename: str, new_tablename: str='', column_case: str='same', progress_handler=None):
+    def copy_table(self, other_db, tablename: str, new_tablename: str='', column_case: str='same', progress_handler=None) -> None:
         '''
         Copy specified table from other easy_db.DataBase to this DB.
         If desired, column names can be set to be all upper or lower-case
@@ -681,7 +696,7 @@ class DataBase():
             print(f'Table "{tablename}" not found.  Table copy aborted.')
             return
 
-        data = other_db.pull_table(tablename, clear_cache=True, progress_handler=progress_handler)  # clearing cache to ensure fresh pull
+        data = other_db.pull(tablename, fresh=True, progress_handler=progress_handler)  # clearing cache to ensure fresh pull
         if column_case.lower() == 'lower':
             columns_and_types = {key.lower(): val for key, val in other_db.columns_and_types(tablename).items()}
             table_data = [{col.lower(): val for col, val in d.items()} for d in data]
@@ -701,6 +716,7 @@ class DataBase():
         if table_data:
             self.append(tablename, table_data)
         print(f'Table {tablename} copied!')
+        self.columns_and_types.cache_clear()
 
 
     def __repr__(self) -> str:
